import os

DB_HOST = os.getenv("DB_HOST", "localhost")
DB_PORT = os.getenv("DB_PORT", 8086)
DB_USER = os.getenv("DB_USER", "admin")
DB_PASSWORD = os.getenv("DB_PASSWORD", "pass123")
DB_NAME = os.getenv("DB_NAME", "PULSAR_DATA_COLLECTION")
DB_PROTOCOL = os.getenv("DB_PROTOCOL", "line")
DB_PREDICTION_MEASURMENT = os.getenv("DB_PREDICTION_MEASURMENT", "prediction")
DB_EVAL_TIMESTAMP_MEASURMENT = os.getenv("DB_EVAL_TIMESTAMP_MEASURMENT", "eval_timestamp")
DB_METRICS_MEASURMENT = os.getenv("DB_METRICS_MEASURMENT", "metrics")



# -*- coding: utf-8 -*-
# import influxdb_client as ifdb
import pandas as pd
from influxdb import DataFrameClient

from .config import DB_NAME


class StorageEngine:
    def __init__(self):
        pass

    def sql_insertion(self, measurment_name, df: pd.DataFrame, database_login=None):

        """
        Function to push Pandas Dataframe into Influx DB.
        """
        client = DataFrameClient(
            database_login.db_host, database_login.db_port, database_login.db_user, database_login.db_password, DB_NAME
        )
        client.create_database(DB_NAME)
        client.get_list_database()
        client.switch_database(DB_NAME)

        client.write_points(df, measurment_name, protocol=database_login.protocol, time_precision="u", tag_columns=("uuid",))

        return df

    def sql_digestion(self, measurment_name, database_login, filters=None):
        client = DataFrameClient(
            database_login.db_host, database_login.db_port, database_login.db_user, database_login.db_password, DB_NAME
        )
        filter_string = ""
        if filters:
            filter_string += "where "
            for key, value in filters.items():
                filter_string += f"{key}{value}"

        client.switch_database(DB_NAME)

        df = client.query(f"select * from {measurment_name} {filter_string}", chunked=True)

        return df



import importlib
import logging
import sys
import uuid
from datetime import datetime
from typing import List, Optional

import numpy as np
import pandas as pd
from pydantic import BaseModel, Field, validator

from ..db_connectors.influxdb.config import (  # DB_NAME,
    DB_EVAL_TIMESTAMP_MEASURMENT,
    DB_HOST,
    DB_METRICS_MEASURMENT,
    DB_PASSWORD,
    DB_PORT,
    DB_PREDICTION_MEASURMENT,
    DB_PROTOCOL,
    DB_USER,
)
from .exceptions import CustomExceptionWithMessage as e

logger = logging.getLogger()

DATABASE_OPERATION_TYPE_INSERT_PREDICTION = "INSERT_PREDICTION"
DATABASE_OPERATION_TYPE_METRICS = "METRICS"

DATABASE_OPERATION_TYPES = (DATABASE_OPERATION_TYPE_INSERT_PREDICTION, DATABASE_OPERATION_TYPE_METRICS)


class DatabaseLogin(BaseModel):
    db_host: str = DB_HOST
    db_port: int = DB_PORT
    db_user: str = DB_USER
    db_password: str = DB_PASSWORD
    protocol: str = DB_PROTOCOL


class DataWithPrediction(BaseModel):
    timestamp: datetime = datetime.utcnow()
    prediction: np.ndarray = Field(...)
    data_points: pd.DataFrame = Field(...)
    features_names: Optional[List[str]]

    class Config:
        arbitrary_types_allowed = True


class DataCaptureParameters(BaseModel):
    operation_type: str = Field(...)
    storage_engine: str
    login_url: Optional[DatabaseLogin]
    model_id: str = ""
    model_version: str = ""
    data_id: str = ""
    y_name: Optional[str]
    pred_name: Optional[str]
    other_labels: Optional[List[str]] = None

    @validator("model_id", "model_version", "data_id")
    def check_model_version(cls, value, values):
        if values.get("operation_type") == DATABASE_OPERATION_TYPE_INSERT_PREDICTION:
            if not value:
                raise ValueError(
                    "Required parameters missing. model_id, model_version, and data_id are required for Insert operation."
                )
            return value
        return ""

    @validator("storage_engine")
    def valid_import(cls, value):
        # Only influxdb is working storage engine rn
        if not value == "influxdb":
            raise e(
                value=value,
                message=f"{value} Storage Engine doesn't exist",
            )
        return value

    @validator("operation_type")
    def validate_op_type(cls, value):
        if value not in DATABASE_OPERATION_TYPES:
            raise e(
                value=value,
                message=f"{value} is an Invalid Operation type",
            )
        return value

    @validator("login_url")
    def validate_login_url(cls, value):
        if not value:
            return DatabaseLogin()
        return value


class DataCapture(DataCaptureParameters):
    def __init__(self, *args, **kwargs):

        return super().__init__(*args, **kwargs)

    def push(self, data=DataWithPrediction):

        """
        Function to convert prediction output to Pandas dataframe
        to be inserted in DB
        """
        if self.operation_type not in (DATABASE_OPERATION_TYPE_INSERT_PREDICTION,):
            raise Exception(f"Method is only allowed for operation type {DATABASE_OPERATION_TYPE_INSERT_PREDICTION}")

        data_with_prediction = data.data_points.copy()
        data_with_prediction["y_pred"] = data.prediction

        data_with_prediction.loc[:, "Timestamp"] = data.timestamp
        data_with_prediction.loc[:, "model_id"] = self.model_id
        data_with_prediction.loc[:, "model_version"] = self.model_version
        data_with_prediction.loc[:, "data_id"] = self.data_id
        if self.pred_name:
            data_with_prediction.loc[:, self.pred_name] = self.pred_name

        data_with_prediction["uuid"] = [uuid.uuid4() for _ in range(len(data_with_prediction.index))]
        # Set 'TimeStamp' field as index of dataframe
        data_with_prediction.set_index("Timestamp", inplace=True)
        DataFactory.sql_ingestion(DB_PREDICTION_MEASURMENT, self.storage_engine, data_with_prediction, self.login_url)
        logger.info("Data was successfully ingested into the db")
        return

    def collect(self, filters: dict = None):
        """Function for retrieving Pandas dataframe from the DB"""
        return DataFactory.sql_digestion(DB_PREDICTION_MEASURMENT, self.storage_engine, self.login_url, filters)

    def collect_eval_timestamp(self):
        """Retrieves last period what was inseted to the database"""
        results = DataFactory.sql_digestion(DB_EVAL_TIMESTAMP_MEASURMENT, self.storage_engine, self.login_url)

        if results and list(results.get("eval_timestamp")):
            df = pd.DataFrame(results.get("eval_timestamp"))
            df["eval_timestamp"] = pd.to_datetime(df["eval_timestamp"])
            return df.iloc[df["eval_timestamp"].argmax()]["eval_timestamp"]
        return None

    def push_eval_timestamp(self, eval_df):
        """Inserts period to the database"""
        DataFactory.sql_ingestion(DB_EVAL_TIMESTAMP_MEASURMENT, self.storage_engine, eval_df, self.login_url)

    def push_metrics(self, metrics_df):
        """Insterts metrics dataframe to the database"""
        DataFactory.sql_ingestion(DB_METRICS_MEASURMENT, self.storage_engine, metrics_df, self.login_url)


class DataFactory:
    @classmethod
    def get_storage_engine(cls, storage_engine: str):
        """Returns storage current storage engine"""
        if storage_engine in ("influxdb",):
            from ..db_connectors.influxdb.db_connection import StorageEngine

            return StorageEngine
        else:
            raise e(
                value=storage_engine,
                message=f"{storage_engine} is an Invalid Storage Engine",
            )

    @classmethod
    def sql_ingestion(cls, measurment_name, storage_engine: str, dataframe: pd.DataFrame, database_login: DatabaseLogin):
        """Function to import DB connection based on storage engine and call sql_insertion"""
        sengine = cls.get_storage_engine(storage_engine)

        sengine().sql_insertion(
            measurment_name,
            df=dataframe,
            database_login=database_login,
        )

    @classmethod
    def sql_digestion(cls, measurment_name, storage_engine: str, database_login: DatabaseLogin = None, filters: dict = None):
        """Function to export DB connection based on storage engine and call sql_digestion"""

        sengine = cls.get_storage_engine(storage_engine)
        return sengine().sql_digestion(measurment_name, database_login=database_login, filters=filters)

    @classmethod
    def imp_module(cls, storage_engine: str):
        """Function to import different DB connections based on StorageEngine passed in input."""

        module = importlib.import_module(storage_engine)
        mod_nm = module.__name__

        if mod_nm in ("influxdb"):
            name = "pulsar_data_collection.db_connectors.influxdb.db_connection"
            if (spec := importlib.util.find_spec(name)) is not None:

                module = importlib.util.module_from_spec(spec)
                sys.modules[name] = module
                spec.loader.exec_module(module)
                cls = getattr(module, "StorageEngine")

            else:
                print(f"can't find the {name!r} module")
        else:
            raise e(
                value=mod_nm,
                message=f"{mod_nm} is an Invalid Storage Engine",
            )

        return cls



# -*- coding: utf-8 -*-
import sqlite3 as db

import pandas as pd


def sql_insertion(df):
    try:
        conn = db.connect("SQLite_Python.db")
        cursor = conn.cursor()
        print("Database created and Successfully Connected to SQLite")

        # cursor.execute("SELECT * FROM mpm_data_ing;")

        df.to_sql("df", conn, if_exists="replace")

        # print(cursor.fetchall())

        # cursor.execute(
        #     """
        #     CREATE TABLE IF NOT EXISTS mpm_data_ing as
        #     SELECT * FROM df
        #     """
        # )

        df.to_sql(name="mpm_13jul", con=conn, if_exists="append", index=False)

        df = pd.read_sql_query("SELECT * from mpm_13jul", conn)
        print(df)
        # cursor.execute("SELECT * FROM mpm_10jul;")
        # print(cursor.fetchall())

    except db.Error as error:
        print("Error while connecting to sqlite", error)
    finally:
        if cursor:
            cursor.close()
            conn.close()
            print("The SQLite connection is closed")

    return df



class CustomExceptionWithMessage(Exception):
    """
    Custom error that is raised in DataCapture submodule
    with FieldValue and Custom Message passed as input
    """

    def __init__(self, value: str, message: str) -> None:
        self.value = value
        self.message = message
        super().__init__(message)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pulsar_data_collection',
 'pulsar_data_collection.data_capture',
 'pulsar_data_collection.db_connectors.influxdb']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.39,<2.0.0',
 'influxdb-client[ciso]>=1.31.0,<2.0.0',
 'influxdb==5.3.1',
 'pandas>=1.4.2,<2.0.0',
 'pydantic>=1.6.2,<2.0.0',
 'pysqlite3>=0.4.7,<0.5.0']

setup_kwargs = {
    'name': 'pulsar-data-collection',
    'version': '0.1.1',
    'description': 'sdk enabling data collection from model serving code for our MPM solution',
    'long_description': '# pulsar_data_collection\n\nPulsar data collection SDK is an open-source Python library for\npushing/processing/collecting features, predictions and metadata. Works with different\ndata storages, at this point InfluxDB is implemented.\n\n## Getting started\n\nInstall Pulsar Data Collection with pip:\n\n```bash\npython3 -m pip install --upgrade pip\npython3 -m pip install --upgrade pulsar-data-collection\n```\n\n### Components\n\nThere are two core components in data collection SDK: storage engine and data capture.\nRight now storage engine implemented only for InfluxDb, it helps to make ingestion and digestion operations\nto the database.\n\n#### Data Capture\n\n`DataCapture` class helps to ingest dataset to database with needed parameters and needed format for future\ndigestion and metrics calculation without any significant changes of data.\n\nIt requires `storage_engine` (available only influxdb right now), `operation_type` (`DATABASE_OPERATION_TYPE_INSERT_PREDICTION`,\n`DATABASE_OPERATION_TYPE_METRICS`),  `login_url` (object of DatabaseLogin class) as input parameters.\n\nOperation type `DATABASE_OPERATION_TYPE_INSERT_PREDICTION` uses for any ingestion operations to the database.\nIt requires additional parameters: `model_id`, `model_version`, `data_id`", `y_name`, `pred_name` what describes\nan input dataset.\nFor operation type `DATABASE_OPERATION_TYPE_METRICS` what commonly uses for retrieving dataset ready for metrics\ncalculation these parameters aren\'t required.\n\nThe last and probably one the most important class to work with is `DataWithPrediction`.\nIt requires two parameters as input: `prediction`, `data_points`. Where `prediction` is prediction value of the model,\nand `data_points` is features dataset. `Push` method of the `DataCapture` takes object of `DataWithPrediction` as\nrequired parameter, and after that makes ingestion operation to database with data transforming, like adding timestamp,\nchanging name of prediction column in dataset, combining features with prediction into single dataset, creating\ninfluxdb unique cache, etc.\n\nList of methods of `DataCapture` class:\n\n- push(data: DataWithPrediction)\n- ingests data to the db after preprocessing it;\n- collect(filters: dict) - retrieves data from db;\n- collect_eval_timestamp - retrieves the newest timestamp in the database;\n- push_eval_timestamp(eval_df: df) - ingesting new one timestamp into db;\n- push_metrics(metrics_df: df) - ingesting metrics dataframe to the database after calculations\n\n### Example usage\n\nInitialize Database credentials:\n\n```python\nfrom pulsar_data_collection.data_capture import DatabaseLogin\ndatabase_login = DatabaseLogin(db_host=<db_host>), db_port=<db_port>, db_user=<db_user>, db_password=<db_password>, protocol=<db_protocol>)\n```\n\nInitialize DataCapture class, depends on operation type use appropriate constant.\nFor inserting data into the database:\n\n```python\nfrom pulsar_data_collection.data_capture import DataCapture, DATABASE_OPERATION_TYPE_INSERT_PREDICTION\n\ndat_predict = DataWithPrediction(prediction=prediction, data_points=to_predict)\n\ndat_capture = DataCapture(\n    storage_engine="influxdb",\n    model_id=<model_id>,\n    model_version=<model_verstion>,\n    data_id=<data_id>,\n    y_name=<y_name>,\n    pred_name=<pred_name>,\n    operation_type=<operation_type>,\n    login_url=<database_login>,\n)\n\ndat_capture.push(dat_predict)\n```\n\nFor collecting data from the database:\n\n```python\nfrom pulsar_data_collection.data_capture import DataCapture, DATABASE_OPERATION_TYPE_METRICS\n\ndat_capture = DataCapture(\n    storage_engine="influxdb",\n    operation_type=DATABASE_OPERATION_TYPE_METRICS,\n    login_url=database_login\n)\n\ndat_capture.collect()\n```\n\nCollection the newest prediction data what wasn\'t precessed\n\n```python\n# receiving the last period of data\n\nlast_eval_timestamp = dat_capture.collect_eval_timestamp()\n\n# if last period exists, collecting only data what wasn\'t collected previously\nif last_eval_timestamp:\n    last_eval_timestamp_str = last_eval_timestamp.strftime(\'%Y-%m-%d %H:%M:%S\')\n    db_df = pd.DataFrame(dat_capture.collect({"time": f">= \'{last_eval_timestamp_str}\'"}).get("prediction"))\nelse:\n    db_df = pd.DataFrame(dat_capture.collect().get("prediction"))\n```\n\nExample of pushing calculated metrics:\n\n```python\ndat_capture.push_metrics(df_result_drift)\n```\n\nExample of pushing the timestamp when metrics were calculated:\n\n```python\ndat_capture.push_eval_timestamp(eval_timestamp_df)\n```\nTODO: add use cases of input dataframes: metrics, prediction, datapoint\n\n## About [PulsarML](https://pulsar.ml/)\n\nPulsarML is a project helping with monitoring your models and gain powerful insights into its performance.\n\nWe released two Open Source packages :\n\n- [pulsar-data-collection](https://github.com/Rocket-Science-Development/pulsar_data_collection) :  lightweight python SDK enabling data collection of features, predictions and metadata from an ML model serving code/micro-service\n- [pulsar-metrics](https://github.com/Rocket-Science-Development/pulsar_metrics) : library for evaluating and monitoring data and concept drift with an extensive set of metrics. It also offers the possibility to use custom metrics defined by the user.\n\nWe also created [pulsar demo](https://github.com/Rocket-Science-Development/pulsar_demo) to display an example use-case showing how to leverage both packages to implement model monitoring and performance management.\n\nWant to interact with the community? join our [slack channel](https://pulsarml.slack.com)\n\nPowered by [Rocket Science Development](https://rocketscience.one/)\n\n## Contributing\n\n1. Fork this repository, develop, and test your changes\n2. open an issue\n3. Submit a pull request with a reference to the issue\n\nTODO: add use cases of input dataframes: metrics, prediction, datapoint\n\n',
    'author': 'Pulsar team',
    'author_email': 'pulsar@data-rs.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Rocket-Science-Development/pulsar_data_collection',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
